/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package uk.co.tp53.NEATG_A;

import java.io.BufferedWriter;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;

/**
 *
 * @author panik
 */
public class CellCache {

    private static List<String> cellsDump;
    private static String fileName;
    private static Path path;
    private final static int SIZE = 1000;

    CellCache(String filePath) {
        cellsDump = new ArrayList<>();
        fileName = filePath + "\\" + UUID.randomUUID() + "eacells.part";
        path = Paths.get(fileName);
        try {
            Files.deleteIfExists(path);
            Files.createFile(path);
        } catch (IOException ex) {
            System.err.format("IOException: %s%n", ex);
        }
    }

    public String getPath() {
        return path.toString();
    }

    public void flush() {
        if (cellsDump.size() > 0) {
            try (BufferedWriter writer = Files.newBufferedWriter(path,
                    Charset.defaultCharset(), StandardOpenOption.APPEND,
                    StandardOpenOption.CREATE)) {
                for (int i = 0; i < cellsDump.size(); i++) {
                    writer.write(cellsDump.get(i));
                    writer.newLine();
                }
                writer.close();
                cellsDump.clear();
            } catch (IOException x) {
                System.err.format("IOException: %s%n", x);
            }
        }
    }

    public void append(String data) {
        cellsDump.add(data);
        if (cellsDump.size() > SIZE) {
            try (BufferedWriter writer = Files.newBufferedWriter(path,
                    Charset.defaultCharset(), StandardOpenOption.APPEND,
                    StandardOpenOption.CREATE)) {
                for (int i = 0; i < cellsDump.size(); i++) {
                    writer.write(cellsDump.get(i));
                    writer.newLine();
                }
                writer.close();
                cellsDump.clear();                
            } catch (IOException x) {
                System.err.format("IOException: %s%n", x);
            }
        }
    }

    public void rename(String fullFileName) {
        File cacheFile = path.toFile();
        File newFile = new File(fullFileName);
        newFile.delete();
        cacheFile.renameTo(newFile);
    }

    public void saveCache(String header, String fullFileName) {
        flush();
        Path finalFile = Paths.get(fullFileName);
        try {
            Files.deleteIfExists(finalFile);
            Files.createFile(finalFile);
            BufferedWriter writer = Files.newBufferedWriter(finalFile,
                    Charset.defaultCharset(), StandardOpenOption.APPEND,
                    StandardOpenOption.CREATE);
            writer.write(header);
            writer.newLine();
            BufferedReader reader = Files.newBufferedReader(path);
            String line = reader.readLine();
            while (line != null) {
                writer.write(line);
                writer.newLine();
                line = reader.readLine();
            }
            writer.close();
            reader.close();
            path.toFile().delete();
        } catch (IOException ex) {
            System.err.format("IOException: %s%n", ex);
        }

    }

    public void clear() {
        try {
            Files.deleteIfExists(path);
//            Files.createFile(path);
        } catch (IOException x) {
            System.err.format("IOException: %s%n", x);
        }
        cellsDump.clear();
    }

    public List<String> readAll() {
        cellsDump = new ArrayList<>();
        try (BufferedReader reader = Files.newBufferedReader(path)) {
            cellsDump = Files.readAllLines(path, Charset.defaultCharset());
            reader.close();
        } catch (IOException x) {
            System.err.format("IOException: %s%n", x);
        }

        return cellsDump;
    }

}
