/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package uk.co.tp53.NEATG_A;

import java.util.*;
import ec.util.MersenneTwisterFast;

/**
 *
 * @author Pan.Pantziarka
 */
public class GridElement {

    public Map<String, NeatCell> eaCells;
    private int runningCount;
    public int maxCellCount;
    private final int optimumCellCount;
    private final int nutrientTarget;
    private final int nutrientRate;
    private int handshakeTolerance;
    private final int maxAge;
    private final double mutationRate;
    private final double invasionRate;
    public String address;
    private static NeatCell normalCell;
    public double[] laggedGeneSupply;
    private int[] neighbours;
    public GridElementStats stats;
    public boolean gridIsNecrotic;
    public boolean needsNewCell;
    public static MersenneTwisterFast r = NeatUtils.getRand();

    GridElement(NeatCell normCell, int add, int maxCount,
            int optimumCount, int handshakeTol) {

        normalCell = normCell;
        gridIsNecrotic = false;
        address = "" + add;
        maxCellCount = maxCount;
        optimumCellCount = optimumCount;
        nutrientTarget = normCell.getNutrientTarget();
        nutrientRate = normCell.getNutrientRate();
        handshakeTolerance = handshakeTol;
        maxAge = normCell.getLifeTime();
        mutationRate = normCell.getMutationRate();
        invasionRate = normCell.getInvasionRate();
        runningCount = optimumCount;
        eaCells = new HashMap<>();
        stats = new GridElementStats(this, normCell.clone());

        for (int i = 0; i < optimumCellCount; i++) {
            NeatCell cell = normCell.clone();
            cell.setClock(1 + r.nextInt(maxAge));
            cell.setKey(address + "_" + i);
            cell.setParentKey("NormCell");
            eaCells.put(cell.getKey(), cell);
        }
        laggedGeneSupply = new double[normCell.getGenomeLength()];
    }

    public void setNeighbours(int[] theHood) {
        neighbours = theHood;
    }

    public int getRandomNeighbour() {
        int n = r.nextInt(neighbours.length);
        return neighbours[n];
    }

    public int cellDensity() {
        return eaCells.size();
    }
    
    public int getMalignantCount() {
        return stats.malignant;
    }

    public int getHandshakeTolerance() {
        return handshakeTolerance;
    }

    public void setHandshakeTolerance(int tol) {
        handshakeTolerance = tol;
    }

    public int canAddCell() {
        //if GE is full return 0 - no more room
        //else return the running count to use as a key for new cell
        int newCount = (cellDensity() - stats.dividing) + 1;
        if (newCount >= maxCellCount) {
            return 0;
        } else {
            //eaCells.put("cell_" + newCount, new NeatCell());
            runningCount++;
            return runningCount;
        }
    }

    public Map<String, NeatCell> clockTick(int nutrientSupply,
            double[] gFactorSupply, IStrategy treatment) {

        Map<String, NeatCell> newUnstressedCells = new HashMap<>();
        Map<String, NeatCell> newStressedCells = new HashMap<>();
        needsNewCell = false;

        //skip this clocktick if the element is necrotic or empty
        if (eaCells.isEmpty() || gridIsNecrotic) {
            needsNewCell = eaCells.isEmpty();  //flag we need some cells
            return new HashMap<>();
        } else {
            //Remove all the cells flagged as to be cleared
            removeDeadCells();
        }

        //Replace dead cells with new ones - depending on grid element status
        int migrateCells = 0;
        for (NeatCell eaCell : eaCells.values()) {
            if (eaCell.divide()) {
                int newKey = canAddCell();  //0=no, >0=yes
                double chanceOfMutation = r.nextDouble();
                if (newKey != 0) {
                    //there's room for another cell
                    //get a new Normal cell or 
                    //if it's Malignant clone and possibly mutate
                    if (eaCell.getType() == CellType.NORMAL) {
                        NeatCell newCell=normalCell.clone();
                        newCell.setKey(address + "_" + newKey);
                        newCell.setState(CellState.HEALTHY); //*************
                        newCell.setClock(); //****************
                        newUnstressedCells.put(newCell.getKey(), newCell);
                    } else {
                        NeatCell newCell = eaCell.clone();
                        newCell.setKey(address + "_" + newKey);
                        newCell.setState(CellState.HEALTHY); //***************
                        newCell.setClock(); //**************
                        if (r.nextBoolean()){ //*****
                            eaCell.setState(CellState.APOPTOTIC); //*******
                        } //*****
//                        eaCell.setState(CellState.HEALTHY); //**************
//                        eaCell.setClock(r.nextInt(eaCell.getLifeTime()));  //********************
                        if (chanceOfMutation < NEATG_A.mutationRate) {
                            newCell.mutate();
                        }
//                        newCell.setState(CellState.HEALTHY);
                        newUnstressedCells.put(newCell.getKey(), newCell);
                    }
                } else {
                    //no more room in this Element, so will need to migrate
                    //if Malignant cell then chance to mutate
                    //Check if there's room
                    migrateCells++;
                    runningCount++;
                    NeatCell stressedCell = eaCell.clone();
                    stressedCell.setKey(address + "_" + runningCount);
                    stressedCell.setState(CellState.HEALTHY);
//                    stressedCell.setClock(); //*************
                    if (stressedCell.getType() == CellType.MALIGNANT) {
//                        eaCell.setState(CellState.HEALTHY);  //************
//                        eaCell.setClock(); //***************
                        if (chanceOfMutation < NEATG_A.mutationRate) {
                            stressedCell.mutate();
                        }
                    }
                    stressedCell.setClock(); //set it to the lifeTime value                    
                    newStressedCells.put(stressedCell.getKey(), stressedCell);
                }
            }
        }

        //Add the new unstressed cells
        if (newUnstressedCells.size() > 0) {
            eaCells.putAll(newUnstressedCells);
            newUnstressedCells.clear();
        }

        //Age population based on nutrients and gene factors
        double[] shareOfFactors
                = treatment.agePopulation(this, nutrientSupply, gFactorSupply);

        //Now to add any stressed  cells - taking account of fitness
        Map<String, NeatCell> migratingCells = new HashMap<>();
        //Put all the stressed cells in as well
        //but some malignant cells get a chance to migrate
        for (NeatCell stressed : newStressedCells.values()) {
            if (stressed.getType() == CellType.NORMAL) {
                eaCells.put(stressed.getKey(), stressed);
//                eaCells.put(address + "_" + runningCount, stressed);
            } else {
                if (r.nextDouble() < NEATG_A.invasionRate) {
                    migratingCells.put(stressed.getKey(), stressed);
//                    migratingCells.put(address + "_" + runningCount,
//                            stressed);
                    migrateCells--;
                } else {
                    eaCells.put(stressed.getKey(), stressed);
//                    eaCells.put(address + "_" + runningCount, stressed);
                }
            }
        }
        
        //Now to remove the least fit cells
        if (migrateCells > 0) {
            flagLeastFitCells(gFactorSupply, migrateCells);
        }

        //compute laggedGeneSupply for next iteration
        laggedGeneSupply = getLaggedGeneFactors(shareOfFactors);

        //return the map of all the cells that are to be migrated
        return migratingCells;
    }

    private int removeDeadCells() {
        //Remove dead cells
        int deadCount = 0;
        Iterator<Map.Entry<String, NeatCell>> iter = eaCells.entrySet().iterator();
        while (iter.hasNext()) {
            Map.Entry<String, NeatCell> cell = iter.next();
            if (cell.getValue().getState() == CellState.TO_BE_CLEARED) {
                iter.remove();
                deadCount++;
            }
        }
        return deadCount;
    }

    public double getNutrientShare() {
        double share = 0;
        for (NeatCell eaCell : eaCells.values()) {
            if (eaCell.getState() == CellState.HEALTHY) {
                share += eaCell.getNutrientTarget();
            }
        }
        return share;
    }

    public int[] getTargets() {
        int[] targets = new int[laggedGeneSupply.length];
        for (NeatCell eaCell : eaCells.values()) {
            if (eaCell.getState() == CellState.HEALTHY) {
                int[] t = eaCell.getGeneTargets();
                for (int i = 0; i < targets.length; i++) {
                    targets[i] += t[i];
                }
            }
        }
        return targets;
    }

    private double[] getLaggedGeneFactors(double[] factorShare) {
        double[] gridGF = new double[factorShare.length];
        for (NeatCell eaCell : eaCells.values()) {
            if (eaCell.getState() == CellState.HEALTHY) {
                gridGF = NeatUtils.sumDblArray(gridGF,
                        eaCell.getGeneExpression(factorShare));
            }
        }
        return gridGF;
    }

    public boolean transformRandomCell() {
        boolean done = false;

        ArrayList<NeatCell> arrayCells;
        arrayCells = new ArrayList<>(eaCells.values());
        int j = r.nextInt(arrayCells.size());
        NeatCell cell = arrayCells.get(j);
        if (cell.getType() == CellType.NORMAL) {
            transformCell(cell);
            done = true;
        }
        return done;

    }

    public void transformCell(NeatCell cell) {
        cell.setType(CellType.MALIGNANT);
        cell.setDefaultMutationRate();
    }

    public void injectCell(String address, NeatCell cell) {
        runningCount++;
        cell.setParentKey(cell.getKey());
        cell.setKey(address + "_" + runningCount);
        eaCells.put(cell.getKey(), cell);
    }

    public NeatCell injectCell(NeatGenome genome) {
        NeatCell cell = new NeatCell(genome, nutrientTarget,
                nutrientRate, mutationRate, invasionRate, maxAge);
        injectCell(address, cell);
        return cell;
    }

    public void setNecrotic() {
        gridIsNecrotic = true;
        for (NeatCell cell : eaCells.values()) {
            cell.setNecrotic();
        }
    }

    public void resetNecrotic() {
        gridIsNecrotic = false;
        for (NeatCell cell : eaCells.values()) {
            cell.resetNecrotic();
        }
    }

    private void flagLeastFitCells(double[] factorSupply, int leastFitCount) {

        int[] targets = getTargets();
        double[] shareOfFactors = new double[targets.length];
        for (int i = 0; i < targets.length; i++) {
            shareOfFactors[i] = factorSupply[i] / targets[i];
        }
        double[] fitnessArray = new double[eaCells.size()];
        int i = 0;
        for (String eaCKey : eaCells.keySet()) {
            NeatCell eaCell = eaCells.get(eaCKey);
            if (eaCell.getState() != CellState.TO_BE_CLEARED) {
                eaCell.updateFitness(shareOfFactors);
                fitnessArray[i] = eaCell.getFitness();
                i++;
            }
        }

        Arrays.sort(fitnessArray);
        for (int j = 0; j < leastFitCount; j++) {
//            double fitVal = fitnessArray[eaCells.size() - (j + 1)];
            double fitVal = fitnessArray[j];
            for (String eaCKey : eaCells.keySet()) {
                NeatCell eaCell = eaCells.get(eaCKey);
                if (eaCell.getFitness() == fitVal) {
                    eaCell.setState(CellState.TO_BE_CLEARED);
                    break;
                }
            }
        }

    }

    public NeatCell getRandomCell() {
        NeatCell cell = new NeatCell();
        if (!eaCells.isEmpty()) {
            ArrayList<NeatCell> arrayCells;
            arrayCells = new ArrayList<>(eaCells.values());
            int j = r.nextInt(arrayCells.size());
            cell = arrayCells.get(j);
        }
        return cell;
    }

    public int handshake(NeatCell cell) {
        //this is where we see if the grid/cell CTI codes align 
        int diff = NeatUtils.getLevenshteinDistance(cell.getCellCTICode(),
                normalCell.getCellCTICode());
//        if (diff < handshakeTolerance) {
//            return 0;
//        } else {
//            return handshakeTolerance-diff;
//        }
        return diff;
    }
}
