package uk.co.tp53.NEATG_A;

/**
 *
 * @author Pan Pantziarka
 */
import java.nio.file.Paths;
import java.util.*;

public class NEATG_A {

    GridElement[][] grid;
    private boolean runnerControlled;
    static NeatScenario sc;
    static int gridWidth;
    static int gridLength;
    static int geneCount;
    static int genLength;
    private static String logType;
    static List<StatsRecord> eaStats;
//    static List<String> cellsDump;
    static CellCache cellsDump;
    static HashMap<String, Double> statsDump;
    static IStrategy defaultStrategy;
    static IStrategy treatmentStrategy;
    static List<String> txtSnapshots;
    static NeatCell normalCell;
    static int runCount;
    static int clockTick;
    static int currentRun;
    static int logPeriod;
    static int maxCellCount;
    static int optimumCellCount;
    static int malignantCellCount;
    static int injectMalignantCellAt;
    static int maxAge;
    static int nutrientTarget;
    static int nutrientRate;
    static int nutrientSupply; // maxCellCount * nutrientRate
    static int malignantNutrientTarget;
    static int malignantNutrientRate;
    static int handshakeTolerance; //Limit of bit diffs between cell and grid
    static int[] normalGeneTarget;
    static int[] normalGeneTolerance;
    static int[] malignantGeneTarget;
    static int[] malignantGeneTolerance;
    static double[] geneFactorSupply; //optimumCells * target
    static double mutationRate;
    static double invasionRate;
    static String name;
    static String filePath;
    static int treatmentTriggerGen;
    static int runnerCode;
    static boolean debug;

    public static void main(String[] args) throws ClassNotFoundException,
            InstantiationException {

        NEATG_A ng = new NEATG_A(args);
        ng.multiRun();
//        ng.dataOutputs(filePath, name);
        ng.tidyUp();
    }

    NEATG_A(String args[]) throws ClassNotFoundException,
            InstantiationException {

        //if running from scenario file use that
        if (args.length == 1) {
            if (args[0].endsWith(".txt")) {
                sc = new NeatScenario(args[0]);
            }
        } else {
            sc = new NeatScenario();
        }

        runnerCode = 0;

        statsDump = new HashMap<>();
//        cellsDump=new CellCache(filePath);
//        cellsDump = new ArrayList<>();

    }

    private void getScenario(NeatScenario sc) {
        gridWidth = 10;
        gridLength = 10;
        genLength = 3000;
        runCount = 1;
        logPeriod = 100;
        maxCellCount = 10;
        optimumCellCount = 4;
        malignantCellCount = 1;
        injectMalignantCellAt = 1;
        maxAge = 50;
        nutrientTarget = 10;
        nutrientRate = 1;
        handshakeTolerance=4;
        nutrientSupply = 10; // maxCellCount * nutrientRate
        geneCount = 5;
        int[] normalGT = {50, 50, 100, 20, 150};
        int[] normalGTol = {10, 10, 20, 5, 30};
        int[] malignantGT = {50, 50, 100, 20, 150};
        int[] malignantGTol = {10, 10, 20, 5, 30};
        double[] gF = {};
        mutationRate = 0.0;
        invasionRate = 0.0;
        logType = "ts";
        name = "Default";
        filePath = "R:\\anticancer.org.uk\\NEATG";
        debug = false;

        //set/get scenario values
        gridWidth = sc.getInt("width", gridWidth);
        gridLength = sc.getInt("length", gridLength);
        genLength = sc.getInt("genLength", genLength);
        logPeriod = sc.getInt("logPeriod", logPeriod);
        runCount = sc.getInt("runCount", runCount);
        maxCellCount = sc.getInt("maxCellCount", maxCellCount);
        optimumCellCount = sc.getInt("optimumCellCount",
                optimumCellCount);
        malignantCellCount = sc.getInt("malignantCellCount",
                malignantCellCount);
        injectMalignantCellAt = sc.getInt("injectMalignantCellAt",
                injectMalignantCellAt);
        maxAge = sc.getInt("maxAge", maxAge);
        handshakeTolerance=sc.getInt("handshakeTolerance", handshakeTolerance);
        nutrientTarget = sc.getInt("nutrientTarget", nutrientTarget);
        nutrientRate = sc.getInt("nutrientRate", nutrientRate);
        malignantNutrientTarget = sc.getInt("malignantNutrientTarget",
                nutrientTarget);
        malignantNutrientRate = sc.getInt("malignantNutrientRate", nutrientRate);
        nutrientSupply = sc.getInt("nutrientSupply", nutrientSupply);
        geneCount = sc.getInt("geneCount", geneCount);
        normalGeneTarget = sc.getIntArray("normalGeneTarget",
                normalGT);
        normalGeneTolerance = sc.getIntArray("normalGeneTolerance",
                normalGTol);
        malignantGeneTarget = sc.getIntArray("malignantGeneTarget",
                malignantGT);
        malignantGeneTolerance = sc.getIntArray("malignantGeneTolerance",
                malignantGTol);
        geneFactorSupply = sc.getDoubleArray("geneFactorSupply", gF);
        mutationRate = sc.getDouble("mutationRate", mutationRate);
        invasionRate = sc.getDouble("invasionRate", invasionRate);
        name = sc.getString("name", name);
        filePath = sc.getString("filePath", filePath);

        logType = sc.getString("logType", logType);
        debug = sc.getBoolean("debug", debug);
        NeatUtils.setDebug(debug);
    }

    public void multiRun() throws ClassNotFoundException,
            InstantiationException {

        //get all the variables for the scenario
        getScenario(sc);

        //Inject an override for this collection of runs
        injectTopLevelOverrides();

        //use the scenario as the schedule source for the default treatment
        String treatmentName = "StrategyNoTreatment";
        defaultStrategy = new StrategyNoTreatment();
        defaultStrategy.setSchedule(sc);

        //Get treatment strategy        
        treatmentName = sc.getString("treatmentName", treatmentName);
        treatmentStrategy = getTreatmentStrategy(sc, treatmentName);
        treatmentStrategy.setSchedule(sc);

        //run experiment runCount times
        for (currentRun = 0; currentRun < runCount; currentRun++) {

            //Create the injected gene factors
            // this is set to the normalTarget value * the optimum cell count
            if (geneFactorSupply.length == 0) {
                geneFactorSupply = new double[geneCount];
                for (int i = 0; i < geneCount; i++) {
                    geneFactorSupply[i] = normalGeneTarget[i]
                            * optimumCellCount;
                }
            }

            //Create normal and malignant genomes
            NeatGenome normalGenome = new NeatGenome(geneCount, maxAge,
                    normalGeneTarget, normalGeneTolerance);
            NeatGenome malignantGenome = new NeatGenome(geneCount, maxAge,
                    malignantGeneTarget, malignantGeneTolerance);

            //Create array of records for detailed stats
            eaStats = new ArrayList<>(genLength + 1);
            
            //Create this instance
            createGrid(normalGenome);

            //Initiaise the array of output data snapshots
            txtSnapshots = new ArrayList<>();

            //now run the scenario on the grid
            singleRun(genLength, nutrientSupply, malignantGenome,
                    geneFactorSupply, logPeriod);

            processOutput();

            //delete objects to make sure it's clean for next iteration
            NeatCell.resetPool();
            statsDump.clear();
            cellsDump.clear();
        }
    }

    public void singleRun(int gens, int nutrientSupply, NeatGenome malignant,
            double[] gFactorSupply, int logPeriod) {

        //Inject an override for this run
        injectRunLevelOverrides();

        //Print the scenario before we start
        System.out.println(NeatUtils.prtRunParams());

        //Loop through entire population of grid elements 
        //for the number of gens specified
        double invasion = 0.0;
        for (clockTick = 0; clockTick < gens + 1; clockTick++) {
            IStrategy treatment = treatmentStrategy;

            if (clockTick == injectMalignantCellAt) {
                //if we're seeding with malignant cells add them
                for (int i = 0; i < malignantCellCount; i++) {
                    GridElement gE = grid[(gridWidth / 2)][(gridLength / 2)];
                    NeatCell cell = gE.injectCell(malignant);
                    cell.setNutrientRate(malignantNutrientRate);
                    cell.setNutrientTarget(malignantNutrientTarget);
                    cell.setParentKey("Orig");
                    gE.transformCell(cell);
                }
            }

            for (GridElement[] gE : grid) {
                for (GridElement gElement : gE) {
                    if (treatmentStrategy.trigger(clockTick, invasion)) {
                        treatment = treatmentStrategy;
                    } else {
                        treatment = defaultStrategy;
                    }
                    Map<String, NeatCell> migratingCells
                            = gElement.clockTick(nutrientSupply, gFactorSupply,
                                    treatment);
                    
                    if (gElement.needsNewCell) {
                        boolean done = false;
                        int attempts = 0;
                        while ((!done) && (attempts < 18)) {
                            int neighbour = gElement.getRandomNeighbour();
                            attempts++;
                            int r = NeatUtils.rowFromAddress(neighbour, gridWidth);
                            int c = NeatUtils.columnFromAddress(neighbour, gridWidth);
                            GridElement nGE = grid[r][c];
                            if (!nGE.eaCells.isEmpty()) {
                                NeatCell cell = nGE.getRandomCell();
                                if (cell.getState() == CellState.HEALTHY) {
                                    gElement.eaCells.put(gElement.address, cell);
                                    done = true;
                                    gElement.needsNewCell = false;
                                }
                            }
                        }
                    } else {
                        for (NeatCell cell : migratingCells.values()) {
                            int neighbour = gElement.getRandomNeighbour();
                            injectCell(neighbour, cell);
                        }
                    }
                    gElement.stats.updateStats(clockTick < injectMalignantCellAt);
                }
            }

            //resetAllNecrotic();
            //flag the grid elements that are necrotic
            String txtMap = NeatUtils.getTxtMap(grid);
            setNecroticGridElements(txtMap);
            txtMap = NeatUtils.getTxtMap(grid);

            //build the stats record for this iteration
            fillStatsDump(currentRun, clockTick, txtMap,
                    treatment.isActive());
            StatsRecord record = getStats(clockTick, txtMap,
                    treatment.isActive());
            invasion = NeatUtils.getPercentInvasion(txtMap);
            eaStats.add(record);

            //output status message if logged iteration 
            if ((clockTick % logPeriod) == 0) {
                logIteration(clockTick, txtMap, record);
            }

        }
    }

    public void createGrid(NeatGenome norm) {

        //Create the normal cell for adding to the grid
        normalCell = new NeatCell(norm, nutrientTarget, nutrientRate,
                mutationRate, invasionRate, maxAge);
        
        //And the cell cache with header and normal cell to start
        cellsDump=new CellCache(filePath);
        
        //Create the rectangular grid
        grid = new GridElement[gridLength][gridWidth];
        
        // Cycle through the grid creating each element
        int address = 0;
        for (int row = 0; row < gridLength; row++) {
            for (int col = 0; col < gridWidth; col++) {
                GridElement gE = new GridElement(normalCell, address,
                        maxCellCount, optimumCellCount, 
                        handshakeTolerance);
                gE.setNeighbours(getNeighbours(address));
                grid[row][col] = gE;
                address++;
            }
        }
    }

    public int transformCell(int row, int col, int count) {
        GridElement gE = grid[row][col];
        int done = 0;
        for (int i = 0; i < count; i++) {
            if (gE.transformRandomCell()) {
                done++;
            }
        }
        return done;
    }

    public int transformCell(int address) {
        int row = NeatUtils.rowFromAddress(address, gridWidth);
        int col = NeatUtils.columnFromAddress(address, gridWidth);

        return transformCell(row, col, 1);
    }

    public void injectCell(int address, NeatCell cell) {
        int row = NeatUtils.rowFromAddress(address, gridWidth);
        int col = NeatUtils.columnFromAddress(address, gridWidth);
        GridElement gE = grid[row][col];
        gE.injectCell("" + address, cell);
    }

    /**
     *
     * @param run - currentRun
     * @param gen - current clockTick
     * @param txtMap - used for gridElement level stats
     * @param treatmentActive - flag to show active treatment
     *
     * statsDump fills a HashMap with all of the data that we can later use in a
     * report generator - data covers entire run Data is stored as a double Keys
     * are constructed as: run.gen.metric.population.variable metrics are count,
     * avg, max, min, count population is A(ll), M(alignant), N(ormal) variable
     * is the name of any variable invasionRate
     */
    public void fillStatsDump(int run, int gen, String txtMap,
            boolean treatmentActive) {

        //Prefix for all keys
        String prefix = runnerCode + ":" + run + ":" + gen + "|";
        //some global run level stats first
        statsDump.put(prefix + "currentRun", (double) run);
        statsDump.put(prefix + "gen", (double) gen);
        statsDump.put(prefix + "treatmentActive", treatmentActive ? 1.0 : 0.0);
        statsDump.put(prefix + "genePool", (double) NeatCell.sizeOfPool());
        statsDump.put(prefix + "handshakeTolerance", (double) handshakeTolerance);

        //then the gridElement distribution
        statsDump.put(prefix + "healthyGridElements", (double) NeatUtils.getGridElementByStatus(txtMap, "H"));
        statsDump.put(prefix + "mixedHealthyGridElements", (double) NeatUtils.getGridElementByStatus(txtMap, "h"));
        statsDump.put(prefix + "mixedTumourGridElements", (double) NeatUtils.getGridElementByStatus(txtMap, "m")
                + NeatUtils.getGridElementByStatus(txtMap, "t"));
        statsDump.put(prefix + "tumourGridElements", (double) NeatUtils.getGridElementByStatus(txtMap, "T"));
        statsDump.put(prefix + "necroticGridElements", (double) NeatUtils.getGridElementByStatus(txtMap, "N"));

        //now the cell level stats...
        int totalCellCount = 0;
        int totalMalignantCellCount = 0;
        int totalNormalCellCount = 0;
        int totalHealthyCellCount = 0;
        int totalHealthyNormalCellCount = 0;
        int totalHealthyMalignantCellCount = 0;
        int totalDividingCellCount = 0;
        int totalApoptoticCellCount = 0;
        int totalNecroticCellCount = 0;
        int totalTumourNormalCellCount = 0;
        int totalMalignantViableCellCount = 0;
        int totalMalignantDividing = 0;
        int totalMalignantApoptotic = 0;
        int totalNormalViable = 0;
//        double sumMutationRate = 0.0;
        double avgMutationRate = 0.0;
        double maxMutationRate = 0.0;
        double minMutationRate = Integer.MAX_VALUE;
//        double sumInvasionRate = 0.0;
        double avgInvasionRate = 0.0;
        double maxInvasionRate = 0.0;
        double minInvasionRate = Integer.MAX_VALUE;
//        double sumDistance = 0.0;
        double avgDistance = 0.0;
        double maxDistance = 0.0;
        double minDistance = Integer.MAX_VALUE;
//        int sumMutationCount = 0;
        double avgMutationCount = 0.0;
        int maxMutationCount = 0;
        int minMutationCount = Integer.MAX_VALUE;
//        int sumLifeTime = 0;
        double avgLifeTime = 0.0;
        int maxLifeTime = 0;
        int minLifeTime = Integer.MAX_VALUE;
//        int sumAgeNormal = 0;
        double avgAgeNormal = 0.0;
//        int sumAgeMalignant = 0;
        double avgAgeMalignant = 0.0;
//        double sumNutrientTarget = 0.0;
        double avgNutrientTarget = 0.0;
        double maxNutrientTarget = 0.0;
        double minNutrientTarget = Double.MAX_VALUE;
//        double sumNutrientRate = 0.0;
        double avgNutrientRate = 0.0;
        double maxNutrientRate = 0.0;
        double minNutrientRate = Double.MAX_VALUE;
//        int[] sumGeneTarget = new int[geneCount];
//        double sumNutrientStore = 0.0;
        double avgNutrientStore = 0.0;
        double maxNutrientStore = 0.0;
        double minNutrientStore = Double.MAX_VALUE;
        double[] avgGeneTarget = new double[geneCount];
        int[] maxGeneTarget = new int[geneCount];
        int[] minGeneTarget = NeatUtils.sumIntArray(new int[geneCount], Integer.MAX_VALUE);
//        int[] sumGeneTolerance = new int[geneCount];
        double[] avgGeneTolerance = new double[geneCount];
        int[] maxGeneTolerance = new int[geneCount];
        int[] minGeneTolerance = NeatUtils.sumIntArray(new int[geneCount], Integer.MAX_VALUE);
        double[] sumGeneFactor = new double[geneCount];
        double[] avgGeneFactor = new double[geneCount];
        double[] maxGeneFactor = new double[geneCount];
        double[] minGeneFactor = NeatUtils.sumDblArray(new double[geneCount], Double.MAX_VALUE);
//        double sumFitnessMalignant = 0.0;
        double avgFitnessMalignant = 0.0;
        double maxFitnessMalignant = 0.0;
        double minFitnessMalignant = Double.MAX_VALUE;
//        double sumFitnessNormal = 0.0;
        double avgFitnessNormal = 0.0;
        double maxFitnessNormal = 0.0;
        double minFitnessNormal = Double.MAX_VALUE;
        Map<NeatCell, Integer> popCensus = new HashMap<>();
        int clonalPopulations = 0;

        for (GridElement[] gE : grid) {
            for (GridElement gElement : gE) {

                //Grid level stats first
                GridElementStats s = gElement.stats;
//                s.updateStats(gen < injectMalignantCellAt);
                totalCellCount += s.totalCellCount; //gElement.cellDensity();
                totalMalignantCellCount += s.malignant;
                totalNormalCellCount += s.normal;
                totalHealthyCellCount += s.healthy;
                totalHealthyNormalCellCount += s.healthyNormal;
                totalHealthyMalignantCellCount += s.healthyMalignant;
                totalMalignantViableCellCount += s.viableMalignant;
                totalMalignantDividing += s.dividingMalignant;
                totalMalignantApoptotic += s.apoptoticMalignant;
                totalNormalViable += s.healthyNormal + s.dividingNormal;
                totalDividingCellCount += s.dividing;
                totalApoptoticCellCount += s.apoptotic;
                totalNecroticCellCount += s.necrotic;

                //Now the tumour-specific values
                if ((s.malignant > 0) && (gen > injectMalignantCellAt)) {
                    //Number of healthy cells in tumour
                    totalTumourNormalCellCount += s.healthyNormal + s.dividingNormal;
                    //Mutation rate
//                    sumMutationRate += s.mutationRate;
                    maxMutationRate = Math.max(maxMutationRate, s.max_mutationRate);
                    minMutationRate = Math.min(minMutationRate, s.min_mutationRate);
                    avgMutationRate += s.avg_mutationRate * s.meanDenomMalignant;
                    //InvasionRate
//                    sumInvasionRate += s.invasionRate;
                    maxInvasionRate = Math.max(maxInvasionRate, s.max_invasionRate);
                    minInvasionRate = Math.min(minInvasionRate, s.min_invasionRate);
                    avgInvasionRate += s.avg_invasionRate * s.meanDenomMalignant;
                    //Distance
//                    sumDistance += s.distance;
                    maxDistance = Math.max(maxDistance, s.max_distance);
                    minDistance = Math.min(minDistance, s.min_distance);
                    avgDistance += s.avg_distance * s.meanDenomMalignant;
                    //MutationCount
//                    sumMutationCount += s.mutationCount;
                    maxMutationCount = Math.max(maxMutationCount, s.max_mutationCount);
                    minMutationCount = Math.min(minMutationCount, s.min_mutationCount);
                    avgMutationCount += s.avg_mutationCount * s.meanDenomMalignant;
                    //Lifetime
//                    sumLifeTime += s.lifeTime;
                    maxLifeTime = Math.max(maxLifeTime, s.max_lifeTime);
                    minLifeTime = Math.min(minLifeTime, s.min_lifeTime);
                    avgLifeTime += s.avg_lifeTime * s.meanDenomMalignant;
                    //Age
                    avgAgeNormal += s.avg_ageNormal * s.meanDenomNormal;
                    avgAgeMalignant += s.avg_ageMalignant * s.meanDenomMalignant;
//                    sumAgeMalignant += s.ageMalignant;
                    //NutrientTarget
//                    sumNutrientTarget += s.nutrientTarget;
                    maxNutrientTarget = Math.max(maxNutrientTarget,
                            s.max_nutrientTarget);
                    minNutrientTarget = Math.min(minNutrientTarget,
                            s.min_nutrientTarget);
                    avgNutrientTarget += s.avg_nutrientTarget * s.meanDenomMalignant;
                    //NutrientRate
//                    sumNutrientRate += s.nutrientRate;
                    maxNutrientRate = Math.max(maxNutrientRate,
                            s.max_nutrientRate);
                    minNutrientRate = Math.min(minNutrientRate,
                            s.min_nutrientRate);
                    avgNutrientRate += s.avg_nutrientRate * s.meanDenomMalignant;
                    //NutrientStore
//                    sumNutrientStore += s.nutrientStore;
                    maxNutrientStore = Math.max(maxNutrientStore,
                            s.max_nutrientStore);
                    minNutrientStore = Math.min(minNutrientStore,
                            s.min_nutrientStore);
                    avgNutrientStore += s.avg_nutrientStore * s.meanDenomMalignant;
                    //FitnessMalignant
//                    sumFitnessMalignant += s.fitnessMalignant;
                    maxFitnessMalignant = Math.max(maxFitnessMalignant,
                            s.max_fitnessMalignant);
                    minFitnessMalignant = Math.min(minFitnessMalignant,
                            s.min_fitnessMalignant);
                    avgFitnessMalignant += s.avg_fitnessMalignant * s.meanDenomMalignant;
                    //GeneTargets & tolerance
                    for (int i = 0; i < geneCount; i++) {
//                        sumGeneTarget[i] += s.geneTargets[i];
                        maxGeneTarget[i] = Math.max(maxGeneTarget[i],
                                s.max_geneTargets[i]);
                        minGeneTarget[i] = Math.min(minGeneTarget[i],
                                s.min_geneTargets[i]);
                        avgGeneTarget[i] += s.avg_geneTargets[i] * s.meanDenomMalignant;
//                    }
//                    //GeneTolerance
//                    for (int i = 0; i < geneCount; i++) {
//                        sumGeneTolerance[i] += s.geneTolerance[i];
                        maxGeneTolerance[i] = Math.max(maxGeneTolerance[i],
                                s.max_geneTolerance[i]);
                        minGeneTolerance[i] = Math.min(minGeneTolerance[i],
                                s.min_geneTolerance[i]);
                        avgGeneTolerance[i] += s.avg_geneTolerance[i] * s.meanDenomMalignant;
                    }
//                    sumAgeNormal += s.ageNormal;
                    //FitnessNormal
//                    sumFitnessNormal += s.fitnessNormal;
                    maxFitnessNormal = Math.max(maxFitnessNormal,
                            s.max_fitnessNormal);
                    minFitnessNormal = Math.min(minFitnessNormal,
                            s.min_fitnessNormal);
                    avgFitnessNormal += s.avg_fitnessNormal * s.meanDenomNormal;
                    //GeneFactor
                    for (int i = 0; i < geneCount; i++) {
                        sumGeneFactor[i] += gElement.laggedGeneSupply[i];
                        maxGeneFactor[i] = getMaxStat(maxGeneFactor[i],
                                gElement.laggedGeneSupply[i], s.meanDenomMalignant + s.meanDenomNormal);
                        minGeneFactor[i] = getMinStat(minGeneFactor[i],
                                gElement.laggedGeneSupply[i], s.meanDenomMalignant + s.meanDenomNormal);
                    }

                    //let's also look at the current cell populations
                    for (NeatCell cell : gElement.eaCells.values()) {
                        if (CellType.MALIGNANT == cell.getType()) {
                            if (!popCensus.containsKey(cell)) {
                                popCensus.put(cell, 1);
                            } else {
                                int currentCount = popCensus.get(cell);
                                popCensus.put(cell, currentCount + 1);
                            }
                        }
                    }

                }

                //Grab these normal stats for when we're looking at non-tumour
                //scenarios
                if ((totalMalignantCellCount == 0) && (gen < injectMalignantCellAt)) {
                    //Age normal
                    avgAgeNormal += s.avg_ageNormal * s.meanDenomNormal;
                    //NutrientStore
//                    sumNutrientStore += s.nutrientStore;
                    maxNutrientStore = Math.max(maxNutrientStore,
                            s.max_nutrientStore);
                    minNutrientStore = Math.min(minNutrientStore,
                            s.min_nutrientStore);
                    avgNutrientStore += s.avg_nutrientStore * s.meanDenomNormal;
                    //FitnessNormal
//                    sumFitnessNormal += s.fitnessNormal;
                    maxFitnessNormal = Math.max(maxFitnessNormal,
                            s.max_fitnessNormal);
                    minFitnessNormal = Math.min(minFitnessNormal,
                            s.min_fitnessNormal);
                    avgFitnessNormal += s.avg_fitnessNormal * s.meanDenomNormal;
                    //GeneFactor
                    for (int i = 0; i < geneCount; i++) {
//                        sumGeneFactor[i] += gElement.laggedGeneSupply[i];
                        maxGeneFactor[i] = getMaxStat(maxGeneFactor[i],
                                gElement.laggedGeneSupply[i], s.meanDenomNormal);
                        minGeneFactor[i] = getMinStat(minGeneFactor[i],
                                gElement.laggedGeneSupply[i], s.meanDenomNormal);
                    }
                }

            }
        }

        clonalPopulations = popCensus.size();

        String p = prefix + "count:A:";
        statsDump.put(p + "totalCellCount", (double) totalCellCount);
        statsDump.put(p + "malignantCellCount", (double) totalMalignantCellCount);
        statsDump.put(p + "normalCellCount", (double) totalNormalCellCount);
        statsDump.put(p + "healthyCellCount", (double) totalHealthyCellCount);
        statsDump.put(p + "dividingCellCount", (double) totalDividingCellCount);
        statsDump.put(p + "apoptoticCellCount", (double) totalApoptoticCellCount);
        statsDump.put(p + "necroticCellCount", (double) totalNecroticCellCount);

        p = prefix + "count:N:";
        statsDump.put(p + "totalTumourNormalCellCount",
                (double) totalTumourNormalCellCount);

        p = prefix + "count:M:";
        statsDump.put(p + "viableCellCount", (double) totalMalignantViableCellCount);
        statsDump.put(p + "healthyCellCount", (double) totalHealthyMalignantCellCount);
        statsDump.put(p + "dividingCellCount", (double) totalMalignantDividing);
        statsDump.put(p + "apoptoticCellCount", (double) totalMalignantApoptotic);
        statsDump.put(p + "clonalPopulations", (double) clonalPopulations);

        p = prefix + "avg:M:";
        if (clonalPopulations > 0) {
            int popSizes = 0;
            for (int v : popCensus.values()) {
                popSizes += v;
            }
            statsDump.put(p + "clonalPopSize", (double) popSizes / popCensus.size());
        } else {
            statsDump.put(p + "clonalPopSize", 0.0);
        }

        //now the mean values - based on selected denominator (viable or healthy
        int meanDenom = 0;
        if ((totalMalignantCellCount == 0) && (gen < injectMalignantCellAt)) {
            //when we've only got normal cells
            if ("viable".equals(GridElementStats.selector)) {
                meanDenom = totalNormalViable;
            } else {
                meanDenom = totalHealthyNormalCellCount;
            }
        } else {
            //when we've got malignant cells
            if ("viable".equals(GridElementStats.selector)) {
                meanDenom = totalMalignantViableCellCount;
            } else {
                meanDenom = totalHealthyMalignantCellCount;
            }
        }

//        avgMutationRate = (double) sumMutationRate / totalMalignantViable; //totalHealthyMalignantCellCount;
        avgMutationRate = avgMutationRate / meanDenom;
        statsDump.put(p + "mutationRate", avgMutationRate);
//        avgInvasionRate = (double) sumInvasionRate / totalMalignantViable; //totalHealthyMalignantCellCount;
        avgInvasionRate = avgInvasionRate / meanDenom;
        statsDump.put(p + "invasionRate", avgInvasionRate);
//        avgDistance = (double) sumDistance / totalMalignantViable; //totalHealthyMalignantCellCount;
        avgDistance = avgDistance / meanDenom;
        statsDump.put(p + "distance", avgDistance);
//        avgMutationCount = (double) sumMutationCount / totalMalignantViable; //totalHealthyMalignantCellCount;
        avgMutationCount = avgMutationCount / meanDenom;
        statsDump.put(p + "mutationCount", avgMutationCount);
//        avgLifeTime = (double) sumLifeTime / totalMalignantViable; //totalHealthyMalignantCellCount;
        avgLifeTime = avgLifeTime / meanDenom;
        statsDump.put(p + "lifeTime", avgLifeTime);
//        avgAgeMalignant = (double) sumAgeMalignant / totalMalignantViable; //totalHealthyMalignantCellCount;
        avgAgeMalignant = avgAgeMalignant / meanDenom;
        statsDump.put(p + "ageMalignant", avgAgeMalignant);
//        avgNutrientTarget = (double) sumNutrientTarget / totalMalignantViable; //totalHealthyMalignantCellCount;
        avgNutrientTarget = avgNutrientTarget / meanDenom;
        statsDump.put(p + "nutrientTarget", avgNutrientTarget);
//        avgNutrientRate = (double) sumNutrientRate / totalMalignantViable; //totalHealthyMalignantCellCount;
        avgNutrientRate = avgNutrientRate / meanDenom;
        statsDump.put(p + "nutrientRate", avgNutrientRate);
//        avgNutrientStore=(double) sumNutrientStore/totalMalignantViable;
        avgNutrientStore = avgNutrientStore / meanDenom;
        statsDump.put(p + "nutrientStore", avgNutrientStore);
//        avgFitnessMalignant = sumFitnessMalignant / totalMalignantViable; //totalHealthyMalignantCellCount;
        avgFitnessMalignant = avgFitnessMalignant / meanDenom;
        statsDump.put(p + "fitnessMalignant", avgFitnessMalignant);
        for (int i = 0; i < geneCount; i++) {
            avgGeneTarget[i]
                    = avgGeneTarget[i] / meanDenom; //totalHealthyMalignantCellCount;
            avgGeneTolerance[i]
                    = avgGeneTolerance[i] / meanDenom; //totalHealthyMalignantCellCount;
            statsDump.put(p + "geneTarget[" + i + "]",
                    avgGeneTarget[i]);
            statsDump.put(p + "geneTolerance[" + i + "]",
                    avgGeneTolerance[i]);
        }

        p = prefix + "max:M:";
        statsDump.put(p + "mutationRate", maxMutationRate);
        statsDump.put(p + "invasionRate", maxInvasionRate);
        statsDump.put(p + "distance", maxDistance);
        statsDump.put(p + "mutationCount", (double) maxMutationCount);
        if (clonalPopulations > 0) {
            statsDump.put(p + "clonalPopSize", (double) Collections.max(popCensus.values()));
        } else {
            statsDump.put(p + "clonalPopSize", 0.0);
        }
        statsDump.put(p + "lifeTime", (double) maxLifeTime);
        statsDump.put(p + "nutrientTarget", maxNutrientTarget);
        statsDump.put(p + "nutrientRate", maxNutrientRate);
        statsDump.put(p + "nutrientStore", maxNutrientStore);
        statsDump.put(p + "fitnessMalignant", maxFitnessMalignant);
        for (int i = 0; i < geneCount; i++) {
            statsDump.put(p + "geneTarget[" + i + "]", (double) maxGeneTarget[i]);
            statsDump.put(p + "geneTolerance[" + i + "]", (double) maxGeneTolerance[i]);
        }

        p = prefix + "min:M:";
        statsDump.put(p + "mutationRate", minMutationRate);
        statsDump.put(p + "invasionRate", minInvasionRate);
        statsDump.put(p + "distance", minDistance);
        statsDump.put(p + "mutationCount", (double) minMutationCount);
        if (clonalPopulations > 0) {
            statsDump.put(p + "clonalPopSize", (double) Collections.min(popCensus.values()));
        } else {
            statsDump.put(p + "clonalPopSize", 0.0);
        }
        statsDump.put(p + "lifeTime", (double) minLifeTime);
        statsDump.put(p + "nutrientTarget", minNutrientTarget);
        statsDump.put(p + "nutrientRate", minNutrientRate);
        statsDump.put(p + "nutrientStore", minNutrientStore);
        statsDump.put(p + "fitnessMalignant", minFitnessMalignant);
        for (int i = 0; i < geneCount; i++) {
            statsDump.put(p + "geneTarget[" + i + "]", (double) minGeneTarget[i]);
            statsDump.put(p + "geneTolerance[" + i + "]", (double) minGeneTolerance[i]);
        }

        p = prefix + "avg:N:";
        int cellCount;
        if (totalMalignantCellCount == 0) {
            cellCount = totalNormalCellCount;
        } else {
            cellCount = totalTumourNormalCellCount;
        }
        avgFitnessNormal = avgFitnessNormal / cellCount;
        statsDump.put(p + "fitnessNormal", avgFitnessNormal);
        avgAgeNormal = avgAgeNormal / cellCount;
        statsDump.put(p + "ageNormal", avgAgeNormal);

        p = prefix + "max:N:";
        statsDump.put(p + "fitnessNormal", maxFitnessNormal);

        p = prefix + "min:N:";
        statsDump.put(p + "fitnessNormal", minFitnessNormal);

        p = prefix + "avg:A:";
        for (int i = 0; i < geneCount; i++) {
            avgGeneFactor[i] = (double) sumGeneFactor[i]
                    / (totalMalignantCellCount + totalNormalCellCount);
            statsDump.put(p + "geneFactor[" + i + "]", avgGeneFactor[i]);
        }

        p = prefix + "max:A:";
        for (int i = 0; i < geneCount; i++) {
            statsDump.put(p + "geneFactor[" + i + "]", maxGeneFactor[i]);
        }

        p = prefix + "min:A:";
        for (int i = 0; i < geneCount; i++) {
            statsDump.put(p + "geneFactor[" + i + "]", minGeneFactor[i]);
        }
    }

    /**
     *
     * @param metric - are count, avg, max, min,
     * @param pop - population is All, Malignant, Normal
     * @param var - the variable name
     * @return double value Data is stored as a double Keys are constructed as:
     * run.gen.metric.population.variable metrics count population is All,
     * Malignant, Normal variable is the name of any variable invasionRate,
     * including index for array variables e.g. geneTarget[i]
     */
    public double getStatDatum(String metric, String pop, String var) {
        String key = runnerCode + ":" + currentRun + ":" + clockTick + "|";
        if (metric.length() > 0) {
            key += metric + ":";
        }
        if (pop.length() > 0) {
            key += pop + ":";
        }
        key += var;
        double val = statsDump.get(key);
        return val;
    }

    private double getMaxStat(double max, double val, int count) {
        if (Double.isNaN(max) && (count > 0)) {
            return val / count;
        } else if (count > 0) {
            return (val / count) > max ? (val / count) : max;
        } else {
            return max;
        }
    }

    private int getMaxStat(int max, int val, int count) {
        if (count > 0) {
            return (val / count) > max ? (val / count) : max;
        } else {
            return max;
        }
    }

    private double getMinStat(double min, double val, int count) {
        if (Double.isNaN(min) && (count > 0)) {
            return val / count;
        } else if (count > 0) {
            return (val / count) < min ? (val / count) : min;
        } else {
            return min;
        }
    }

    private int getMinStat(int min, int val, int count) {
        if (count > 0) {
            return (val / count) < min ? (val / count) : min;
        } else {
            return min;
        }
    }

    public StatsRecord getStats(int gen, String txtMap,
            boolean treatmentActive) {
        double[] factors = new double[geneCount];
        double[] targets = new double[geneCount];
        double[] tolerance = new double[geneCount];
        StatsRecord record = new StatsRecord(geneCount);
        record.setGen((int) getStatDatum("", "", "gen"));
        record.setPoolSize((int) getStatDatum("", "", "genePool"));
        record.setAvgMutationRate(getStatDatum("avg", "M", "mutationRate"));
        record.setAvgInvasionRate(getStatDatum("avg", "M", "invasionRate"));
        record.setAvgDistance(getStatDatum("avg", "M", "distance"));
        record.setAvgMutationCount(getStatDatum("avg", "M", "mutationCount"));
        record.setAvgLifeTime(getStatDatum("avg", "M", "lifeTime"));
        record.setMaxLifeTime(getStatDatum("max", "M", "lifeTime"));
        record.setAvgNutrientTarget(getStatDatum("avg", "M", "nutrientTarget"));
        record.setAvgNutrientRate(getStatDatum("avg", "M", "nutrientRate"));
        record.setAvgNutrientStore(getStatDatum("avg", "M", "nutrientStore"));
        record.setHandshakeTolerance((int) getStatDatum("","","handshakeTolerance"));

//        int totalCellCount = 0;
//        int totalMalignantCells = 0;
//        int totalNormalCells = 0;
//        int totalHealthyCells = 0;
        double[] tmpFactors = new double[geneCount];
        for (GridElement[] gE : grid) {
            for (GridElement gElement : gE) {
//                totalHealthyCells += gElement.stats.healthy;
//                totalCellCount += gElement.cellDensity();
//                totalMalignantCells += gElement.stats.malignant;
//                totalNormalCells += gElement.stats.normal;
                for (int i = 0; i < geneCount; i++) {
                    double factor = gElement.laggedGeneSupply[i];
                    if (Double.isInfinite(factor)) {
                        factor = 0;
                    }
                    tmpFactors[i] += factor;
                }
            }
        }

        for (int i = 0; i < geneCount; i++) {
//            factors[i] = factors[i] / totalHealthyCells;
            targets[i] = getStatDatum("avg", "M", "geneTarget[" + i + "]");
            tolerance[i] = getStatDatum("avg", "M", "geneTolerance[" + i + "]");
            factors[i] = getStatDatum("avg", "A", "geneFactor[" + i + "]");
        }

//        tempResults = getFitnessMalignant();
        record.setAvgMalignantFitness(getStatDatum("avg", "M", "fitnessMalignant"));
        record.setMaxMalignantFitness(getStatDatum("max", "M", "fitnessMalignant"));
        record.setAvgNormalFitness(getStatDatum("avg", "N", "fitnessNormal"));
        record.setAvgGeneFactors(factors);
        record.setAvgGeneTargets(targets);
        record.setAvgGeneTolerance(tolerance);
        record.setTotalCellCount((int) getStatDatum("count", "A", "totalCellCount"));
        record.setTotalMalignantCells((int) getStatDatum("count", "A", "malignantCellCount"));
        record.setTotalNormalCells((int) getStatDatum("count", "A", "normalCellCount"));
        record.setTotalHealthyCells((int) getStatDatum("count", "A", "healthyCellCount"));
        record.setHealthyGrid((int) getStatDatum("", "", "healthyGridElements"));
        record.setMixedHealthyGrid((int) getStatDatum("", "", "mixedHealthyGridElements"));
        record.setMixedTumourGrid((int) getStatDatum("", "", "mixedTumourGridElements"));
        record.setTumourGrid((int) getStatDatum("", "", "tumourGridElements"));
        record.setNecroticGrid((int) getStatDatum("", "", "necroticGridElements"));
        record.setTreatmentActive((getStatDatum("", "", "treatmentActive") == 1.0));

        return record;
    }

    private int[] getNeighbours(int address) {
        ArrayList<Integer> theTempHood = new ArrayList<>();

        int row = NeatUtils.rowFromAddress(address, gridWidth);
        int col = NeatUtils.columnFromAddress(address, gridWidth);

        for (int iCol = - 1; iCol < 2; iCol++) {
            for (int jRow = -1; jRow < 2; jRow++) {
                int newCol = iCol + col;
                if ((newCol > -1) && (newCol < gridWidth)) {
                    int newRow = jRow + row;
                    if ((newRow > -1) && (newRow < gridLength)) {
                        int x = (newRow * gridWidth) + newCol;
                        if (x != address) {
                            theTempHood.add(new Integer(x));
                        }
                    }
                }
            }
        }

        int j = 0;
        for (Integer i : theTempHood) {
            if (i != -1) {
                j++;
            }
        }
        int[] theHood = new int[j];
        j = 0;
        for (Integer i : theTempHood) {
            if (i != -1) {
                theHood[j] = i;
                j++;
            }
        }
        Arrays.sort(theHood);
        return theHood;
    }

    private String setNecroticGridElements(String mapOfGrid) {

        int rows = gridLength;
        int cols = gridWidth;
        int bufferLayer = 3; //((1 + gridLength) / 10);
        StringBuilder newMap = new StringBuilder(
                mapOfGrid.replaceAll(System.lineSeparator(), ""));

        for (int row = 0; row < rows; row++) {
            for (int col = 0; col < cols; col++) {
                processGridMap(newMap, col, row, bufferLayer);
            }
        }

        return newMap.toString();
    }

    private static IStrategy getTreatmentStrategy(NeatScenario sc,
            String treatmentName)
            throws InstantiationException, ClassNotFoundException {

        String treatmentSchedule = "this";
        treatmentTriggerGen = 2500;
        double treatmentTriggerInvasion = 0;

        treatmentSchedule = sc.getString("treatmentSchedule", treatmentSchedule);
        treatmentTriggerGen = sc.getInt("treatmentTriggerGen",
                treatmentTriggerGen);
        treatmentTriggerInvasion = sc.getDouble("treatmentTriggerInvasion",
                treatmentTriggerInvasion);

        //Assume the default before looking for real
        IStrategy treatment = defaultStrategy;

        //get the treatment class
        if (!treatmentName.equals("none")) {
            try {
                Object strategy = Class.
                        forName("uk.co.tp53.NEATG_A." + treatmentName)
                        .newInstance();
                treatment = (IStrategy) strategy;
                treatment.setGen(treatmentTriggerGen);
                treatment.setInvasion(treatmentTriggerInvasion);
                if (treatmentSchedule.equals("this")) {
                    treatment.setSchedule(sc);
                } else {
                    treatment.setSchedule(
                            new NeatScenario(treatmentSchedule));
                }
                return treatment;
            } catch (IllegalAccessException ex) {
                System.out.println(ex.toString());
            }
        }
        return treatment;
    }

    public void processOutput() {

        //dump all stats
        if (logType.contains("s")) {
            System.out.println(eaStats.get(0).describe());
            for (StatsRecord eaStat : eaStats) {
                System.out.println(eaStat.toString());
            }
            System.out.println("\n\n");
        }
        
        //and the rest for a single run...
        dataOutputs(filePath, name);
        
    }

    public void dataOutputs(String filePath, String name) {
        //dump the data to file - if file path is given
        if (filePath.length() != 0) {
            String filePrefix = sc.getString("filePrefix", name);
            if (filePath.equals("this")) {
                filePath = Paths.get("").toAbsolutePath().toString();
            }
            String outputType = sc.getString("outputType", "overwrite");
            String fileName;
            //txt snapshots first
            fileName = NeatUtils.getOutputFileName(outputType, filePath,
                    filePrefix, ".eatxt");
            NeatUtils.outputToFile(txtSnapshots, fileName);
            //now the stats records
            fileName = NeatUtils.getOutputFileName(outputType, filePath,
                    filePrefix, ".eastats");
            String header = eaStats.get(0).describe();
            NeatUtils.outputToFile(eaStats, header, fileName);
            //and the scenario listing with all the fields
            fileName = NeatUtils.getOutputFileName(outputType, filePath,
                    filePrefix, ".ea");
            NeatUtils.outputToFile(sc.toStringArray(), fileName);
            //dump the cells at the end of the process for verbose logging
            if (logType.contains("v")) {
                fileName = NeatUtils.getOutputFileName(outputType, filePath,
                        filePrefix, ".eacells");
                header = NeatUtils.allCellsHeader(normalCell, geneCount,
                        geneCount);
                header=header.replace(":--:--:--:--:", ":--:--:"); //hack for single run
                cellsDump.saveCache(header, fileName);
//                cellsDump.prepend(header);
//                NeatUtils.outputToFile(new StringBuilder(header), fileName);
//                cellsDump.rename(fileName);
//                NeatUtils.appendToFile(cellsDump, fileName);
            }
        }
    }

    public void logIteration(int tick, String txt, StatsRecord record) {
        String out = NeatUtils.statusOutput(tick, txt, record);
        System.out.println(out);
        txtSnapshots.add(out);
        //Now we want the cells...
        NeatUtils.dumpAllCells(grid, tick, cellsDump);
    }

    public void resetAllNecrotic() {
        for (GridElement[] gE : grid) {
            for (GridElement gElement : gE) {
                gElement.resetNecrotic();
            }
        }
    }

    private void processGridMap(StringBuilder newMap, int col, int row,
            int bufferLayer) {
        char current = newMap.charAt(col + (row * gridWidth));
        if ((current == 'T') || (current == 'N')) {
            boolean necrotic = true;
            for (int r = row - bufferLayer;
                    r < row + bufferLayer + 1; r++) {
                for (int c = col - bufferLayer;
                        c < col + bufferLayer + 1; c++) {
                    if ((r >= 0) && (r < gridLength)
                            && (c >= 0) && (c < gridWidth)) {
                        char neighbour = newMap.charAt(c + (r * gridWidth));
//                        if ((neighbour != 'T') && (neighbour != 'N')) {
                        if ((neighbour == 'H') || (neighbour == 'h')
                                || (neighbour == 'm') || (neighbour == 't')) { //|| (neighbour =='n')
                            necrotic = false;
                            r = gridLength + 1;
                            c = gridWidth + 1;
                        }
                    }
                }
            }
            if (necrotic) {
                grid[row][col].setNecrotic();
//                newMap.setCharAt(col + (row * gridWidth), 'N');
            } else {
                if (current == 'N') {
                    grid[row][col].resetNecrotic();
//                    newMap.setCharAt(col + (row * gridWidth), 'T');
                }
            }
        }
    }

    public void setRunnerControlled(boolean rC) {
        runnerControlled = rC;
    }

    public boolean getRunnerControlled() {
        return runnerControlled;
    }

    public void tidyUp() {
        statsDump.clear();
//        cellsDump.clear();
    }

    /**
     * This is a chance to inject variables at the run level
     *
     * @param args
     */
    public static void injectRunLevelOverrides() {
    }

    //This is a chance to inject variables at the runCount level    
    public static void injectTopLevelOverrides() {
    }
}
