/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor. 
 */
package uk.co.tp53.NEATG_A;

import ec.util.MersenneTwisterFast;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 *
 * @author Pan.Pantziarka
 */
public class NeatCell {

    private CellType cellType;
    private CellState cellState;
    private CellState preNecroticState;
    private NeatGenome genome;
    private int cellClock;
    private int cumulativeMutationCount;
    private double nutrientStore;
    private int nutrientRate;
    private int nutrientTarget;
    private double fitness;
    private int lifeTime;
    private double mutationRate;
    private double invasionRate;
    private String key;
    private String parentKey;
    final public static double inActiveFlag = 1e-23;
    private static MersenneTwisterFast r
            = NeatUtils.getRand();
    private static Map<String, Integer> cellPool = new LinkedHashMap<>();

    NeatCell() {
        lifeTime = 100;
        cellClock = lifeTime;
        nutrientTarget = 25;
        nutrientStore = nutrientTarget;
        nutrientRate = 1;
        cellType = CellType.NORMAL;
        cellState = CellState.HEALTHY;
        fitness = 0;
        cumulativeMutationCount = 0;
        mutationRate = 0.0;
        invasionRate = 0.0;
    }

    NeatCell(NeatGenome norm, int nutrientTarget, int nutrientRate,
            double mutRate, double prolRate, int maxAge) {
        this();
        this.lifeTime = maxAge;
        this.nutrientRate = nutrientRate;
        this.nutrientTarget = nutrientTarget;
        this.nutrientStore = nutrientTarget;
        this.mutationRate = mutRate;
        this.invasionRate = prolRate;
        this.cellClock = maxAge; //1 + r.nextInt(maxAge);
        this.genome = norm.clone();
        addToPool(this);
    }

    NeatCell(NeatCell inCell) {
        this();
        lifeTime = inCell.getLifeTime();
        cellType = inCell.getType();
        cellState = inCell.getState();
        cellClock = inCell.getClock();
        nutrientRate = inCell.getNutrientRate();
        nutrientTarget = inCell.getNutrientTarget();
        nutrientStore = inCell.getNutrientStore();
        mutationRate = inCell.getMutationRate();
        invasionRate = inCell.getInvasionRate();
        cumulativeMutationCount = inCell.getCumulativeMutationCount();
        genome = inCell.genome.clone();
        key=inCell.getKey();
        parentKey=inCell.getParentKey();
        addToPool(this);
    }

    @Override
    public NeatCell clone() {
        NeatCell cell = new NeatCell(this);
        cell.nutrientStore = cell.nutrientTarget;
        cell.parentKey=cell.key;
        return cell;
    }

    public void feedCell(double nutrient) {
        nutrientStore += (nutrient - nutrientRate);
        if (nutrientStore < 0) {
            cellClock = 0;
        } else if (cellClock > 0) {
            cellClock--;
        }
    }

    public int getGenomeLength() {
        return genome.genes.length;
    }

    public int getGeneTarget(int i) {
        return genome.genes[i].getTarget();
    }

    public int getGeneTolerance(int i) {
        return genome.genes[i].getTolerance();
    }

    public int getCumulativeMutationCount() {
        return cumulativeMutationCount;
    }

    public void processFactors(double[] shareOfFactor) {
        int notHealthyCount = 0;
        double actualFactor;
        for (int i = 0; i < shareOfFactor.length; i++) {
            if (shareOfFactor[i] != inActiveFlag) {
                NeatGene gene = genome.genes[i];
                actualFactor = shareOfFactor[i] * gene.getTarget();
                if (!gene.health(actualFactor)) {
                    notHealthyCount++;
                }
            }
        }
        while (notHealthyCount > 0) {
            if ((cellClock > 0) && (cellState == CellState.HEALTHY)) {
                cellClock--;
            }
            notHealthyCount--;
        }
    }

    public void changeState() {
        //assume no params means grid is not necrotic
        changeState(false);
    }

    public void changeState(boolean necroticGrid) {
        if ((cellClock == 0) || (nutrientStore < 0)) {
            switch (cellState) {
                case HEALTHY:
                    cellState = CellState.DIVIDING;
                    break;
                case DIVIDING:
                    if (cellType == CellType.NORMAL) {
                        cellState = CellState.APOPTOTIC;
                    } else if (nutrientStore < 0) {
                        cellState = CellState.APOPTOTIC;
                    } else if (!necroticGrid) {
                        cellState = CellState.HEALTHY;
                        cellClock = lifeTime;
                    }
                    break;
                case APOPTOTIC:
                    cellState = CellState.TO_BE_CLEARED;
                    break;
                case NECROTIC:
                    if (!necroticGrid) {
                        cellState = CellState.TO_BE_CLEARED;
                    }
                    break;
            }
        }
    }

    //Added handshake code
    public void changeState(boolean necroticGrid, int handshake, int tol) {
        if ((cellClock == 0) || (nutrientStore < 0)) {
            switch (cellState) {
                case HEALTHY:
                    cellState = CellState.DIVIDING;
                    break;
                case DIVIDING:
                    if ((nutrientStore < 0) || (cellType == CellType.NORMAL)) {
                        cellState = CellState.APOPTOTIC;
                    } else if (!necroticGrid) {
                        if (tol==0) { 
                            cellState = CellState.APOPTOTIC;
                        } else {
                            int chance=r.nextInt(1+handshake);
                            if (tol>chance){
                                cellState = CellState.HEALTHY;
                                cellClock = lifeTime;
                            } else {
                                cellState=CellState.APOPTOTIC;
                            }
                        }
                    }
                    break;
                case APOPTOTIC:
                    cellState = CellState.TO_BE_CLEARED;
                    break;
                case NECROTIC:
                    if (!necroticGrid) {
                        cellState = CellState.TO_BE_CLEARED;
                    }
                    break;
            }
        }
    }

    public CellType getType() {
        return cellType;
    }

    public CellState getState() {
        return cellState;
    }

    public int getClock() {
        return cellClock;
    }

    public void setClock(int age) {
        cellClock = age;
    }

    public void setClock() {
        cellClock = lifeTime;
    }

    public void setRandomClock() {
        cellClock = 1 + r.nextInt(lifeTime);
    }

    public double getNutrientStore() {
        return nutrientStore;
    }

    public int getNutrientRate() {
        return nutrientRate;
    }

    public void setNecrotic() {
        preNecroticState = cellState;
        cellState = CellState.NECROTIC;
    }

    public void resetNecrotic() {
        if (preNecroticState != null) {
            cellState = preNecroticState;
        }
    }

    public int getNutrientTarget() {
        return nutrientTarget;
    }
    
    public String getKey() {
        return key;
    }
    
    public String getParentKey() {
        return parentKey;
    }
    
    public void setKey(String k) {
        key=k;
    }
    
    public void setParentKey(String k) {
        parentKey=k;
    }

    public boolean divide(boolean gridIsNecrotic) {
        if (gridIsNecrotic) {
            return false;
        } else if ((cellState == CellState.DIVIDING)
                && (nutrientStore >= nutrientTarget)) {
            return true;
        } else {
            return false;
        }
    }

    public boolean divide() {
        //assume grid has room to expand
        return divide(false);
    }

    public double getFitness() {
        if (fitness > 1.0) {
            System.out.println("Fitness exceeded value of 1.0");
        }
        return fitness;
    }

    public void setNutrientTarget(int target) {
        nutrientTarget = target;
    }

    public void setNutrientRate(int rate) {
        nutrientRate = rate;
    }

    public void setNutrientStore(int store) {
        nutrientStore = store;
    }

    public void setType(CellType type) {
        cellType = type;
    }

    public void setState(CellState state) {
        cellState = state;
    }

    public void setCumulativeMutationCount(int count) {
        cumulativeMutationCount = count;
    }

    public NeatGenome getGenome() {
        return genome;
    }

    public int[] getGeneTargets() {
        int[] targets = new int[genome.genes.length];
        for (NeatGene gene : genome.genes) {
            int i = gene.getGeneID();
            targets[i] = gene.getTarget();
        }
        return targets;
    }

    public double[] getGeneExpression(double[] factors) {
        double[] expression = new double[factors.length];
        for (NeatGene gene : genome.genes) {
            int i = gene.getGeneID();
            if (factors[i] == NeatCell.inActiveFlag) {
                expression[i] = NeatCell.inActiveFlag;
            } else {
                expression[i] = gene.expression(factors[i]);
            }
        }
        return expression;
    }

    public void updateFitness(double[] factors) {
        if ((nutrientStore < 0) || (cellClock == 0)) {
            fitness = 0;
        } else {
            double fitness2 = Math.exp(-genome.getFitness(factors));
            double fitness1 = 1 - Math.exp(-(nutrientStore - nutrientTarget) / cellClock);

            //added to handle NaN more gracefully
            if (Double.isNaN(fitness2) || (fitness2 > 1.0)) {
                fitness2 = 0;
            }
            fitness = (fitness1 + fitness2) / 2; //fitness2;
            if (fitness < 0.0) {
                fitness = 0;
            }
        }
    }

    @Override
    public String toString() {
        return getType() + ":" + getState() + ":" + lifeTime + ":"
                + cumulativeMutationCount + ":"
                + cellClock + ":"
                + mutationRate + ":" + invasionRate + ":"
                + nutrientStore + ":" + nutrientRate + ":"
                + nutrientTarget + ":" + genome;

    }

    public static String describe(int count) {
        return "Type:State:LifeTime:MutCount:Clock:Mutation:Invasion:"
                + "N Store:N Rate:N Target:"
                + NeatGenome.describe(count);
    }

    public void mutate() {
        int segment = r.nextInt(5 + this.getGenomeLength());
        cumulativeMutationCount++;
        switch (segment) {
            case 0:
                nutrientRate = NeatUtils.mutateInt(nutrientRate);
                if (nutrientRate == 0) {
                    nutrientRate = 1;
                }
                break;
            case 1:
                nutrientTarget = NeatUtils.mutateInt(nutrientTarget);
                if (nutrientTarget == 0) {
                    nutrientTarget = 1;
                }
                break;
            case 2:
                if (r.nextBoolean()) {
                    lifeTime += 1 + r.nextInt(this.getGenomeLength());
                } else {
                    lifeTime -= 1 + r.nextInt(this.getGenomeLength());
                    if (lifeTime <= 0) {
                        lifeTime = 1;
                    }
                }
                break;
            case 3: {
                double change = (mutationRate / 10)
                        * (1 + r.nextInt(this.getGenomeLength()));
                if (r.nextBoolean()) {
                    if ((mutationRate + change) > 1.0) {
                        mutationRate = 1.0;
                    } else {
                        mutationRate += change;
                    }
                } else if ((mutationRate - change) < 0.0) {
                    mutationRate = NEATG_A.mutationRate / 10;
                } else {
                    mutationRate -= change;
                }
                break;
            }
            case 4: {
                double change = (invasionRate / 10)
                        * (1 + r.nextInt(this.getGenomeLength()));
                if (r.nextBoolean()) {
                    if ((invasionRate + change) > 1.0) {
                        invasionRate = 1.0;
                    } else {
                        invasionRate += change;
                    }
                } else if ((invasionRate - change) < 0.0) {
                    invasionRate = NEATG_A.invasionRate / 10;
                } else {
                    invasionRate -= change;
                }
                break;
            }
            default:
                genome.mutate();
                break;
        }
        addToPool(this);
    }

    public double getMutationRate() {
        return mutationRate;
    }

    public void setDefaultMutationRate() {
        mutationRate = NEATG_A.mutationRate;
    }

    public void setMutationRate(double newRate) {
        mutationRate = newRate;
    }

    public double getInvasionRate() {
        return invasionRate;
    }

    public void setDefaultInvasionRate() {
        invasionRate = NEATG_A.invasionRate;
    }

    public void setInvasionRate(double newRate) {
        invasionRate = newRate;
    }

    @Override
    public int hashCode() {
        String keyVals = lifeTime + "|" + mutationRate + "|" + invasionRate
                + "|" + nutrientRate + "|" + nutrientTarget + "|" + genome;
        return keyVals.hashCode();
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final NeatCell other = (NeatCell) obj;
        if (!this.genome.equals(other.genome)) {
            return false;
        }
        if (this.nutrientRate != other.nutrientRate) {
            return false;
        }
        if (this.nutrientTarget != other.nutrientTarget) {
            return false;
        }
        if (this.invasionRate != other.invasionRate) {
            return false;
        }
        if (this.lifeTime != other.lifeTime) {
            return false;
        }
        return true;
    }

    /**
     * @return the lifeTime
     */
    public int getLifeTime() {
        return lifeTime;
    }

    /**
     * @param lifeTime the lifeTime to set
     */
    public void setLifeTime(int lifeTime) {
        this.lifeTime = lifeTime;
    }

    public boolean isViable() {
        return isHealthy() || (cellState == CellState.DIVIDING);
    }

    public boolean isHealthy() {
        return (cellState == CellState.HEALTHY);
    }

    public double getDistance(NeatCell otherCell) {
        double deltaLifeTime = Math.pow((this.lifeTime - otherCell.lifeTime), 2.0);
        double deltaNutrientTarget = Math.pow(
                (this.nutrientTarget - otherCell.nutrientTarget), 2.0);
        double deltaNutrientRate = Math.pow(
                (this.nutrientRate - otherCell.nutrientRate), 2.0);
        double deltaMutationRate = Math.pow(
                (this.mutationRate - otherCell.mutationRate), 2.0);
        double deltaInvasionRate = Math.pow((this.invasionRate - otherCell.invasionRate), 2.0);
        double deltaGenome = this.genome.getDistance(otherCell.genome);
        double distance = deltaLifeTime + deltaNutrientTarget
                + deltaMutationRate + deltaInvasionRate + deltaNutrientRate
                + deltaGenome;
        return Math.sqrt(distance);
    }

    public String getCellCTICode() {
        StringBuilder CTI = new StringBuilder();
        for (int i = 0; i < this.getGenomeLength(); i++) {
            CTI.append(this.getGeneTarget(i));
            CTI.append(this.getGeneTolerance(i));
        }
        CTI.append(this.getLifeTime());
        CTI.append(this.getNutrientTarget());
        CTI.append(this.getNutrientRate());
        CTI.append(("" + this.getInvasionRate()).replace(".", ""));
        CTI.append(("" + this.getMutationRate()).replace(".", ""));
        return CTI.toString();
    }

    public static void addToPool(NeatCell cell) {
        String k = "k" + cell.hashCode();
        if (!cellPool.containsKey(k)) {
            cellPool.put(k, 1);
        } else {
            int currentCount = cellPool.get(k);
            cellPool.put(k, currentCount + 1);
        }
    }

    public static int sizeOfPool() {
        return cellPool.size();
    }

    public static void resetPool() {
        cellPool = null;
        cellPool = new LinkedHashMap<>();
    }

}
