/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package uk.co.tp53.NEATG_A;

import java.io.*;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author panik
 */
public class NeatRunner extends NEATG_A {

    public static String varName;
    public static String varType;
    public static int varIntStart;
    public static int varIntStep;
    public static int varIntIterations;
    public static int varIntCurrentValue;
    public static double varDoubleStart;
    public static double varDoubleStep;
    public static double varDoubleCurrentValue;
    public static double[] varDblArrayStart;
    public static double[] varDblArrayCurrentValue;
    private static NeatScenario runSc;
    private static int runCode;
    private static String dbFile;
    private static String cellFile;
    public static boolean customDbFormat;
    public static String runnerScenarioFile;
    public static ArrayList<String> fieldNamesAll;
    public static ArrayList<String> fldNamesAlias;
    public static ArrayList<String> fldNamesOrder;

    public static void main(String args[]) throws ClassNotFoundException,
            InstantiationException {

        //if running from runner file use that
        if (args.length == 1) {
            if (args[0].endsWith("runner.txt")) {
                runnerScenarioFile = args[0];
                runSc = new NeatScenario(runnerScenarioFile);
            }
        } else {
            runSc = new NeatScenario();
        }

        //Fiddle about to get args[] for calling NEATG_A
        ArrayList<String> argList = new ArrayList<>();
        String scenarioFile = runSc.getString("scenarioFile", "");
        argList.add(scenarioFile);
        String[] runArgs = new String[argList.size()];
        runArgs = argList.toArray(runArgs);

        //Create the NEATG_A instance
        NeatRunner runner = new NeatRunner(runArgs);

        varName = runSc.getString("<variableName>", "");
        if (varName.length() > 0) {
            varType = runSc.getString("<variableType>", "int");
            if (varType.equals("int")) {
                varIntStart = runSc.getInt("<variableStart>", 0);
                varIntCurrentValue = varIntStart;
                varIntStep = runSc.getInt("<variableStep>", 1);
            } else if (varType.startsWith("dblArray")) {
                int length = Integer.parseInt(varType.split("\\|")[1]);
                varDblArrayStart = runSc.getDoubleArray("<variableStart>",
                        new double[length]);
                varDblArrayCurrentValue = varDblArrayStart.clone();
                varType = "dblArray";
                varIntStep = runSc.getInt("<variableStep>", 1);
            } else if (varType.equals("double")) {
                varDoubleStart = runSc.getDouble("<variableStart>", 0.0);
                varDoubleStep = runSc.getDouble("<variableStep>", 0.0);
                varDoubleCurrentValue = varDoubleStart;
            }
            varIntIterations = runSc.getInt("<variableIterations>", 0);
        }

        runCode = 0;
        for (int i = 0; i < varIntIterations; i++) {
            //Now to inject the variables into NEATG_A
            NeatRunner.injectTopLevelOverrides();
            //OK to perform the run
            runner.multiRun();
//            runner.processOutput();
            if (varType.equals("int")) {
                varIntCurrentValue += varIntStep;
            } else if (varType.equals("dblArray")) {
                for (int j = 0; j < varDblArrayStart.length; j++) {
                    varDblArrayCurrentValue[j] += varIntStep;
                }
            } else if (varType.equals("double")) {
                varDoubleCurrentValue += varDoubleStep;
            }
            runCode++;
        }

//        //Process the outputs..
//        runner.processOutput();
        //Now clear up anything we need to...
        runner.tidyUp();

    }

    NeatRunner(String[] ScenarioName) throws ClassNotFoundException,
            InstantiationException {

        super(ScenarioName);
        super.setRunnerControlled(true);

    }

    public static void injectTopLevelOverrides() {

        //Let's cycle through the props to get the fixed variables  
        for (Enumeration<Object> e = runSc.props.keys(); e.hasMoreElements();) {
            String key = e.nextElement().toString();
            String[] keyChunks = key.split(">");
            if (key.startsWith("<fixed>")) {
                String var = keyChunks[1];
                String val = runSc.props.getProperty(key);
                NEATG_A.sc.setProp(var, val);
            }
        }

        //Now for the variable variables
        if (varType.equals("int")) {
            NEATG_A.sc.setProp(varName, Integer.toString(varIntCurrentValue));
        } else if (varType.equals("dblArray")) {
            String v = "";
            for (int j = 0; j < varDblArrayCurrentValue.length; j++) {
                v += varDblArrayCurrentValue[j] + "|";
            }
            v = v.substring(0, v.lastIndexOf("|"));
            NEATG_A.sc.setProp(varName, v);
        } else if (varType.equals("double")) {
            NEATG_A.sc.setProp(varName, Double.toString(varDoubleCurrentValue));
        }

        //And the runnerCode - needed to identify which experiment this is
        NEATG_A.runnerCode = runCode;

    }

    @Override
    public void processOutput() {

        String lf = System.lineSeparator();
        StringBuilder record = new StringBuilder();
        boolean firstRun = (runCode == 0) && (currentRun == 0);

        if (firstRun) {
            //just dump the stats record from the last run
            //dataOutputs(filePath, name); 
            System.out.println("Experiments " + varIntIterations);
            System.out.println("Runs per experiment " + runCount);
            System.out.println("Unique data points " + statsDump.size());

            //Get output file name
            dbFile = runSc.getString("dbFile", "");
            customDbFormat = runSc.getBoolean("customDbFormat", false);
            cellFile = runSc.getString("cellFile", "");

            //Let's get the keyset to extract the full set of field names
            fieldNamesAll = new ArrayList<>();
            int nameStart;
            String name;
            for (String key : statsDump.keySet()) {
                if (key.startsWith("0:0:")) {
                    nameStart = key.lastIndexOf("|");
                    name = key.substring(nameStart + 1);
                    if (!fieldNamesAll.contains(name)) {
                        fieldNamesAll.add(name);
                    }
                }
            }

            //Now check for the custom properties and ordering
            fldNamesAlias = new ArrayList<>();
            if (customDbFormat) {
                try {
                    for (String key : Files.readAllLines(Paths.get(runnerScenarioFile))) {
                        if (key.startsWith("<repFld")) {
                            String val = key.split(">")[1];
                            fldNamesAlias.add(val);
                        }
                    }
                } catch (IOException ex) {
                    Logger.getLogger(NeatRunner.class.getName()).log(Level.SEVERE, null, ex);
                }
//            Collections.sort(fldNamesAlias);
            } else {
                fldNamesAlias = fieldNamesAll;
            }

            //
            fldNamesOrder = new ArrayList<>();
            if (customDbFormat) {
                for (int i = 0; i < fldNamesAlias.size(); i++) {
                    String field = fldNamesAlias.get(i);
                    String[] names = field.split("\\|");
                    if (fieldNamesAll.contains(names[0])) {
                        fldNamesOrder.add(names[0]);
                        if (names[1].length() > 0) {
                            fldNamesAlias.set(i, names[1]);
                        }
                    }
                }
            } else {
                fldNamesOrder = fieldNamesAll;
            }

            //Let's get the db record header for the first time through
            record.append("Experiment|Run|Generation");
            for (String field : fldNamesAlias) {
                record.append("|");
                record.append(field);
            }
            record.append(lf);
            System.out.println(record);

        }

        //Use field name and length variables to generate each db record
        int i = runCode;
        int j = currentRun;
        for (int k = 0; k < (genLength + 1); k++) {
            record.append(i).append("|").append(j).
                    append("|").append(k);
            for (String field : fldNamesOrder) {
                String key = i + ":" + j + ":" + k + "|" + field;
                record.append("|");
                record.append(cleanVal(statsDump.get(key)));
            }
            record.append(lf);
        }

        if (dbFile.length() > 0) {
            if (firstRun) {
                NeatUtils.outputToFile(record, dbFile);
            } else {
                NeatUtils.appendToFile(record, dbFile);
            }
        }
        System.out.println(record);

        //Now the eacells data
        if (cellFile.length() > 0) {
            cellsDump.flush();
            if (firstRun) {
                String header = NeatUtils.allCellsHeader(normalCell,
                        geneCount, geneCount);
                header = "RunCode:Experiment:" + header+lf;
                NeatUtils.outputToFile(new StringBuilder(header), cellFile);
            }

            if (currentRun < 2) {  //Simply because of file size issues
                Path path = Paths.get(cellsDump.getPath());
                String line;
                try (BufferedWriter writer = Files.newBufferedWriter(
                        Paths.get(cellFile), StandardOpenOption.APPEND);
                        BufferedReader reader = Files.newBufferedReader(path)) {
                    line = reader.readLine();
                    while (line != null) {
                        writer.write(runCode + ":" + currentRun + ":" + line);
                        writer.newLine();
                        line = reader.readLine();
                    }
                    writer.close();
                    reader.close();
                    cellsDump.clear();
                } catch (IOException x) {
                    System.err.format("IOException: %s%n", x);
                }
            }
        }

    }

    @Override
    public void logIteration(int tick, String txt, StatsRecord record) {
//        String out = NeatUtils.statusOutput(tick, txt, record);
        System.out.println("Progress : runCode " + runCode + ": "
                + " currentRun : " + currentRun
                + " clockTick : " + tick);
        NeatUtils.dumpAllCells(grid, tick, cellsDump);
    }

    private String cleanVal(double val) {
        if (Double.isNaN(val)
                || Double.isInfinite(val)
                || (val == Double.MAX_VALUE)
                || (val == Integer.MAX_VALUE)) {
            return "-";
        } else {
            return Double.toString(val);
        }
    }

}
