/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package uk.co.tp53.NEATG_A;

import ec.util.MersenneTwisterFast;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author panik
 */
public class NeatUtils {

    public static boolean debug;

    private NeatUtils() {
        debug = false;

    }

    public static void setDebug(boolean status) {
        debug = status;
    }

    public static String getTxtMap(GridElement[][] grid) {
        String txtOutput = "";
        for (GridElement[] gE : grid) {
            for (GridElement gElement : gE) {
                txtOutput += gElement.stats.txtPrint();
            }
            txtOutput += System.lineSeparator();
        }
        return txtOutput;
    }

    public static void appendToFile(StringBuilder output, String fullFileName) {
        Path path = Paths.get(fullFileName);
        try {
            BufferedWriter writer;
            writer = Files.newBufferedWriter(path, Charset.defaultCharset(),
                    StandardOpenOption.APPEND);
            writer.append(output);
            writer.close();

        } catch (IOException ex) {
            Logger.getLogger(NEATG_A.class
                    .getName()).log(Level.SEVERE, null, ex);
        }
    }

    public static void appendToFile(List output, String fullFileName) {
        Path path = Paths.get(fullFileName);
        try {
            BufferedWriter writer;
            writer = Files.newBufferedWriter(path, Charset.defaultCharset(),
                    StandardOpenOption.APPEND);
            for (int i = 0; i < output.size(); i++) {
                writer.write(output.get(i).toString());
                writer.newLine();
            }
            writer.close();

        } catch (IOException ex) {
            Logger.getLogger(NEATG_A.class
                    .getName()).log(Level.SEVERE, null, ex);
        }
    }

    public static void outputToFile(List output, String header,
            String fullFileName) {
        Path path = Paths.get(fullFileName);
        try {
            BufferedWriter writer;
            Files.deleteIfExists(path);
            writer = Files.newBufferedWriter(path, Charset.defaultCharset(),
                    StandardOpenOption.CREATE);
            if (header.length() > 0) {
                writer.write(header.toCharArray());
                writer.newLine();
            }
            for (int i = 0; i < output.size(); i++) {
                writer.write(output.get(i).toString());
                writer.newLine();
            }
            writer.close();

        } catch (IOException ex) {
            Logger.getLogger(NEATG_A.class
                    .getName()).log(Level.SEVERE, null, ex);
        }
    }

    public static void outputToFile(StringBuilder output, String header,
            String fullFileName) {
        Path path = Paths.get(fullFileName);
        try {
            BufferedWriter writer;
            Files.deleteIfExists(path);
            writer = Files.newBufferedWriter(path, Charset.defaultCharset(),
                    StandardOpenOption.CREATE);
            if (header.length() > 0) {
                writer.write(header);
                writer.newLine();
            }
            writer.append(output);
//            for (int i = 0; i < output.size(); i++) {
//                writer.write(output.get(i).toString());
//                writer.newLine();
//            }
            writer.close();

        } catch (IOException ex) {
            Logger.getLogger(NEATG_A.class
                    .getName()).log(Level.SEVERE, null, ex);
        }
    }

    public static void outputToFile(List output, String fullFileName) {
        //Empty header string
        outputToFile(output, "", fullFileName);
    }

    public static void outputToFile(StringBuilder output, String fullFileName) {
        //Empty header string
        outputToFile(output, "", fullFileName);
    }

    public static String getOutputFileName(String outType, String path,
            String prefix, String ext) {
        String fileName = path + File.separator + prefix + ext;
        if (outType.equals("append")) {
            int i = 0;
            while (Files.exists(Paths.get(fileName))) {
                fileName = path + File.separator + prefix + "_" + i + ext;
                i++;
            }
        }
        return fileName;
    }

    public static String allCellsHeader(NeatCell normalCell, int geneCount,
            int cellCount) {
        StringBuilder allCells = new StringBuilder();

        //Header
        allCells.append("Gen:");
        allCells.append("Key:");
        allCells.append("Parent:");
        allCells.append("Fitness:");
        allCells.append(NeatCell.describe(geneCount));
        allCells.append("Address:");
        allCells.append("Density:");
        allCells.append("Malignant:");
        allCells.append("Hash:");
        allCells.append("MetHash:");
        allCells.append(System.lineSeparator());

        //add the normal cell first
        allCells.append(0); //Gen
        allCells.append(":--:--:--:--:--:"); //Key:Fitness
        allCells.append(normalCell.toString()); //:Description
        allCells.append("--:"); //Address
        allCells.append(0); //Density
        allCells.append(":");
        allCells.append("0:"); //Malignant
        allCells.append("k").append(normalCell.hashCode()); //Hash
        allCells.append(":");
        String metHash = normalCell.getLifeTime() + "|" + normalCell.getNutrientRate()
                + "|" + normalCell.getNutrientTarget();
        allCells.append("k").append(metHash.hashCode()); //MetHash
        allCells.append(":");
//        allCells.append(System.lineSeparator());

        return allCells.toString();

    }

    public static void dumpAllCells(GridElement[][] grid, int currentGen,
            CellCache cells) {
        StringBuilder allCells;
        String metHash;
        NeatCell cell;

        for (GridElement[] gE : grid) {
            for (GridElement gElement : gE) {
                if (gElement.stats.malignant > 0) {
                    for (String cellKey : gElement.eaCells.keySet()) {
                        cell = gElement.eaCells.get(cellKey);
                        if (cell.getType() != CellType.NORMAL) {
                            allCells = new StringBuilder();
                            allCells.append(currentGen); //Gen
                            allCells.append(":");
                            allCells.append(cellKey); //Key
                            allCells.append(":");
                            allCells.append(cell.getParentKey());
                            allCells.append(":");
                            allCells.append(cell.getFitness()); //Fitness
                            allCells.append(":");
                            allCells.append(cell.toString()); //Description
                            allCells.append(gElement.address); //Address
                            allCells.append(":");
                            allCells.append(gElement.cellDensity()); //Density
                            allCells.append(":");
                            allCells.append(gElement.getMalignantCount()); //MC
                            allCells.append(":");                            
                            allCells.append("k").append(cell.hashCode());
                            allCells.append(":");
                            metHash = cell.getLifeTime() + "|"
                                    + cell.getNutrientRate()
                                    + "|" + cell.getNutrientTarget();
                            allCells.append("k").append(metHash.hashCode());
                            allCells.append(":");
//                            allCells.append(System.lineSeparator());
                            cells.append(allCells.toString());
//                            cells.add(allCells.toString());
                        }
                    }
                }
            }
        }

    }

    public static int getGridElementByStatus(String txtMap, String flag) {
        return txtMap.length() - txtMap.replace(flag, "").length();
    }

    public static double getPercentInvasion(String txtMap) {
        int necrotic = NeatUtils.getGridElementByStatus(txtMap, "N");
        int tumour = NeatUtils.getGridElementByStatus(txtMap, "T");

        return 100 * (double) (necrotic + tumour)
                / txtMap.replace(System.lineSeparator(), "").length();
    }

    public static String prtGridGeneFactors(double[] factors) {
        String s = "";
        for (int i = 0; i < factors.length; i++) {
            s += "Factor " + i + " avg value=" + factors[i] + " ";
        }
        return s;
    }

    public static int rowFromAddress(int address, int width) {
        int row = address / width;
        return row;
    }

    public static int columnFromAddress(int address, int width) {
        int col = address % width;
        return col;
    }

    public static MersenneTwisterFast getRand() {
        if (debug == false) {
            return new MersenneTwisterFast(System.currentTimeMillis());
        } else {
            return new MersenneTwisterFast(2323);
        }
    }

    public static int mutateInt(int start) {
        MersenneTwisterFast rand
                = getRand();
        StringBuilder s_x
                = new StringBuilder("0" + Integer.toBinaryString(start));
        int i = rand.nextInt(s_x.length());
        if (s_x.charAt(i) == '0') {
            s_x.setCharAt(i, '1');
        } else {
            s_x.setCharAt(i, '0');
        }
        return Integer.parseInt(s_x.toString(), 2);
    }

    public static String statusOutput(int clockTick, String txtOutput,
            StatsRecord record) {
        String l = System.lineSeparator();
        String output = "Generation: " + clockTick + l;
        output += txtOutput + System.lineSeparator();
        output += prtGridGeneFactors(record.getAvgGeneFactors()) + l;
        output += String.format("Avg mutation rate %f%% %n",
                record.getAvgMutationRate());
        output += String.format("Avg invasion rate %f%% %n",
                record.getAvgInvasionRate());
        output += ("Peak fitness malignant "
                + record.getMaxMalignantFitness() + l);
        output += ("Average fitness malignant "
                + record.getAvgMalignantFitness() + l);
        output += ("Total no. malignant cells "
                + record.getTotalMalignantCells() + l);
        output += ("Total no. active cells "
                + record.getTotalHealthyCells() + l);
        output += ("No. of elements all Tumour: "
                + record.getTumourGrid() + l);
        output += ("No. of elements all Necrotic: "
                + record.getNecroticGrid() + l);
        output += ("Percent invasion "
                + getPercentInvasion(txtOutput) + l);
        output += l;
        return output;

    }

    public static double[] sumDblArray(double[] ar1, double[] ar2) {
        double[] aRes = new double[ar1.length];
        for (int i = 0; i < aRes.length; i++) {
            aRes[i] = ar1[i] + ar2[i];
        }
        return aRes;
    }

    public static double[] sumDblArray(double[] ar1, double ar2) {
        double[] aRes = new double[ar1.length];
        for (int i = 0; i < aRes.length; i++) {
            aRes[i] = ar1[i] + ar2;
        }
        return aRes;
    }

    public static int[] sumIntArray(int[] ar1, int ar2) {
        int[] aRes = new int[ar1.length];
        for (int i = 0; i < aRes.length; i++) {
            aRes[i] = ar1[i] + ar2;
        }
        return aRes;
    }

    public static String prtRunParams() {
        String l = System.lineSeparator();
        String out = "Scenario : " + NEATG_A.name + l;
        out += "Grid width: " + NEATG_A.gridWidth + l;
        out += "Grid length: " + NEATG_A.gridLength + l;
        out += "Gene count : " + NEATG_A.geneCount + l;
        out += "Generations : " + NEATG_A.genLength + l;
        out += "Run count : " + NEATG_A.runCount + l;
        out += "Log period : " + NEATG_A.logPeriod + l;
        out += "Max cell count : " + NEATG_A.maxCellCount + l;
        out += "Optimum cell count : " + NEATG_A.optimumCellCount + l;
        out += "Malignant cell count : " + NEATG_A.malignantCellCount + l;
        out += "Max age : " + NEATG_A.maxAge + l;
        out += "Nutrient target : " + NEATG_A.nutrientTarget + l;
        out += "Nutrient rate : " + NEATG_A.nutrientRate + l;
        out += "Nutrient supply : " + NEATG_A.nutrientSupply + l;
        out += "File path : " + NEATG_A.filePath + l;
        out += "Mutation rate : " + NEATG_A.mutationRate + l;
        out += "Invasion rate : " + NEATG_A.invasionRate + l;
        out += "Normal gene targets : "
                + Arrays.toString(NEATG_A.normalGeneTarget) + l;
        out += "Normal gene tolerance : "
                + Arrays.toString(NEATG_A.normalGeneTolerance) + l;
        out += "Malignant gene targets : "
                + Arrays.toString(NEATG_A.malignantGeneTarget) + l;
        out += "Malignant gene tolerance : "
                + Arrays.toString(NEATG_A.malignantGeneTolerance) + l;
        out += "Gene factor supply : "
                + Arrays.toString(NEATG_A.geneFactorSupply) + l;

        return out;
    }

    /*
        Code from org.apache.commons.text.similarity - LevenshteinDistance.java
     */
    public static int getLevenshteinDistance(CharSequence left,
            CharSequence right) {

        if (left == null || right == null) {
            throw new IllegalArgumentException("CharSequences must not be null");
        }

        /*
           This implementation use two variable to record the previous cost counts,
           So this implementation use less memory than previous impl.
         */
        int n = left.length(); // length of left
        int m = right.length(); // length of right

        if (n == 0) {
            return m;
        } else if (m == 0) {
            return n;
        }

        if (n > m) {
            // swap the input strings to consume less memory
            final CharSequence tmp = left;
            left = right;
            right = tmp;
            n = m;
            m = right.length();
        }

        final int[] p = new int[n + 1];

        // indexes into strings left and right
        int i; // iterates through left
        int j; // iterates through right
        int upperLeft;
        int upper;

        char rightJ; // jth character of right
        int cost; // cost

        for (i = 0; i <= n; i++) {
            p[i] = i;
        }

        for (j = 1; j <= m; j++) {
            upperLeft = p[0];
            rightJ = right.charAt(j - 1);
            p[0] = j;

            for (i = 1; i <= n; i++) {
                upper = p[i];
                cost = left.charAt(i - 1) == rightJ ? 0 : 1;
                // minimum of cell to the left+1, to the top+1, diagonally left and up +cost
                p[i] = Math.min(Math.min(p[i - 1] + 1, p[i] + 1), upperLeft + cost);
                upperLeft = upper;
            }
        }

        return p[n];
    }

}
